@@ -8,6 +8,7 @@
 #include <gtk/gtk.h>
 #include <string.h>
 #include <glib/gi18n.h>
+#include <math.h>
 
 #include "callbacks.h"
 #include "links.h"
@@ -48,67 +49,90 @@ cb_buffer_changed(girara_session_t* session)
   }
 }
 
+static void
+update_visible_pages(zathura_t* zathura) {
+  const unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
+
+  for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
+    zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
+    GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
+    ZathuraPage* zathura_page_widget = ZATHURA_PAGE(page_widget);
+
+    if (page_is_visible(zathura->document, page_id) == true) {
+      /* make page visible */
+      if (zathura_page_get_visibility(page) == false) {
+        zathura_page_set_visibility(page, true);
+        zathura_page_widget_update_view_time(zathura_page_widget);
+        zathura_renderer_page_cache_add(zathura->sync.render_thread, zathura_page_get_index(page));
+      }
+
+    } else {
+      /* make page invisible */
+      if (zathura_page_get_visibility(page) == true) {
+        zathura_page_set_visibility(page, false);
+        /* If a page becomes invisible, abort the render request. */
+        zathura_page_widget_abort_render_request(zathura_page_widget);
+      }
+
+      /* reset current search result */
+      girara_list_t* results = NULL;
+      g_object_get(page_widget, "search-results", &results, NULL);
+      if (results != NULL) {
+        g_object_set(page_widget, "search-current", 0, NULL);
+      }
+    }
+  }
+}
+
 void
-cb_view_vadjustment_value_changed(GtkAdjustment* GIRARA_UNUSED(adjustment), gpointer data)
+cb_view_hadjustment_value_changed(GtkAdjustment* adjustment, gpointer data)
 {
   zathura_t* zathura = data;
-  if (zathura == NULL || zathura->document == NULL || zathura->ui.page_widget == NULL) {
+  if (zathura == NULL || zathura->document == NULL) {
     return;
   }
 
-  GtkAdjustment* view_vadjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-  GtkAdjustment* view_hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-
-  /* current adjustment values */
-  GdkRectangle view_rect = {
-    .x      = 0,
-    .y      = 0,
-    .width  = gtk_adjustment_get_page_size(view_hadjustment),
-    .height = gtk_adjustment_get_page_size(view_vadjustment)
-  };
-
-  int page_padding = 1;
-  girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
-
-  GdkRectangle center = {
-    .x      = (view_rect.width  + 1) / 2,
-    .y      = (view_rect.height + 1) / 2,
-    .width  = (2 * page_padding) + 1,
-    .height = (2 * page_padding) + 1
-  };
-
-  unsigned int number_of_pages = zathura_document_get_number_of_pages(zathura->document);
-  double scale = zathura_document_get_scale(zathura->document);
-
-  bool updated = false;
-  /* find page that fits */
-  for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
-    zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
-
-    GdkRectangle page_rect = {
-      .width  = zathura_page_get_width(page)  * scale,
-      .height = zathura_page_get_height(page) * scale
-    };
-    GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
-    gtk_widget_translate_coordinates(page_widget,
-                                     zathura->ui.session->gtk.view, 0, 0, &page_rect.x, &page_rect.y);
-
-    if (gdk_rectangle_intersect(&view_rect, &page_rect, NULL) == TRUE) {
-      if (zathura_page_get_visibility(page) == false) {
-	zathura_page_set_visibility(page, true);
-	zathura_page_widget_update_view_time(ZATHURA_PAGE(page_widget));
-	zathura_page_cache_add(zathura, zathura_page_get_index(page));
-      }
-      if (zathura->global.update_page_number == true && updated == false
-          && gdk_rectangle_intersect(&center, &page_rect, NULL) == TRUE) {
-        zathura_document_set_current_page_number(zathura->document, page_id);
-        updated = true;
-      }
-    } else {
-      zathura_page_set_visibility(page, false);
-    }
+  /* Do nothing in index mode */
+  if (girara_mode_get(zathura->ui.session) == zathura->modes.index) {
+    return;
   }
 
+  update_visible_pages(zathura);
+
+  double position_x = zathura_adjustment_get_ratio(adjustment);
+  double position_y = zathura_document_get_position_y(zathura->document);
+  unsigned int page_id = position_to_page_number(zathura->document, position_x, position_y);
+
+  zathura_document_set_position_x(zathura->document, position_x);
+  zathura_document_set_position_y(zathura->document, position_y);
+  zathura_document_set_current_page_number(zathura->document, page_id);
+
+  statusbar_page_number_update(zathura);
+}
+
+void
+cb_view_vadjustment_value_changed(GtkAdjustment* adjustment, gpointer data)
+{
+  zathura_t* zathura = data;
+  if (zathura == NULL || zathura->document == NULL) {
+    return;
+  }
+
+  /* Do nothing in index mode */
+  if (girara_mode_get(zathura->ui.session) == zathura->modes.index) {
+    return;
+  }
+
+  update_visible_pages(zathura);
+
+  double position_x = zathura_document_get_position_x(zathura->document);
+  double position_y = zathura_adjustment_get_ratio(adjustment);
+  unsigned int page_id = position_to_page_number(zathura->document, position_x, position_y);
+
+  zathura_document_set_position_x(zathura->document, position_x);
+  zathura_document_set_position_y(zathura->document, position_y);
+  zathura_document_set_current_page_number(zathura->document, page_id);
+
   statusbar_page_number_update(zathura);
 }
 
@@ -121,29 +145,23 @@ cb_view_hadjustment_changed(GtkAdjustment* adjustment, gpointer data)
   zathura_adjust_mode_t adjust_mode =
     zathura_document_get_adjust_mode(zathura->document);
 
-  gdouble lower, upper, page_size, value, ratio;
-  bool zoom_center = false;
-
-  switch (adjust_mode) {
-    center:
-    case ZATHURA_ADJUST_BESTFIT:
-    case ZATHURA_ADJUST_WIDTH:
-      lower = gtk_adjustment_get_lower(adjustment);
-      upper = gtk_adjustment_get_upper(adjustment);
-      page_size = gtk_adjustment_get_page_size(adjustment);
-      value = ((upper - lower) - page_size) / 2.0;
-      zathura_adjustment_set_value(adjustment, value);
-      break;
-    default:
-      girara_setting_get(zathura->ui.session, "zoom-center", &zoom_center);
-      if (zoom_center) {
-        goto center;
-      }
-
-      ratio = zathura_adjustment_get_ratio(zathura->ui.hadjustment);
-      zathura_adjustment_set_value_from_ratio(adjustment, ratio);
-      break;
+  /* Do nothing in index mode */
+  if (girara_mode_get(zathura->ui.session) == zathura->modes.index) {
+    return;
   }
+
+  /* Don't scroll we're focusing the inputbar. */
+  if (adjust_mode == ZATHURA_ADJUST_INPUTBAR) {
+    return;
+  }
+
+  /* save the viewport size */
+  unsigned int view_width = (unsigned int)floor(gtk_adjustment_get_page_size(adjustment));
+  zathura_document_set_viewport_width(zathura->document, view_width);
+
+  /* reset the adjustment, in case bounds have changed */
+  double ratio = zathura_document_get_position_x(zathura->document);
+  zathura_adjustment_set_value_from_ratio(adjustment, ratio);
 }
 
 void
@@ -155,93 +173,75 @@ cb_view_vadjustment_changed(GtkAdjustment* adjustment, gpointer data)
   zathura_adjust_mode_t adjust_mode =
     zathura_document_get_adjust_mode(zathura->document);
 
+  /* Do nothing in index mode */
+  if (girara_mode_get(zathura->ui.session) == zathura->modes.index) {
+    return;
+  }
+
   /* Don't scroll we're focusing the inputbar. */
   if (adjust_mode == ZATHURA_ADJUST_INPUTBAR) {
     return;
   }
 
-  double ratio = zathura_adjustment_get_ratio(zathura->ui.vadjustment);
+  /* save the viewport size */
+  unsigned int view_height = (unsigned int)floor(gtk_adjustment_get_page_size(adjustment));
+  zathura_document_set_viewport_height(zathura->document, view_height);
+
+  /* reset the adjustment, in case bounds have changed */
+  double ratio = zathura_document_get_position_y(zathura->document);
   zathura_adjustment_set_value_from_ratio(adjustment, ratio);
 }
 
 void
-cb_adjustment_track_value(GtkAdjustment* adjustment, gpointer data)
+cb_refresh_view(GtkWidget* GIRARA_UNUSED(view), gpointer data)
 {
-  GtkAdjustment* tracker = data;
-
-  gdouble lower = gtk_adjustment_get_lower(adjustment);
-  gdouble upper = gtk_adjustment_get_upper(adjustment);
-  if (lower != gtk_adjustment_get_lower(tracker) ||
-      upper != gtk_adjustment_get_upper(tracker)) {
+  zathura_t* zathura = data;
+  if (zathura == NULL || zathura->document == NULL) {
     return;
   }
 
-  gdouble value = gtk_adjustment_get_value(adjustment);
-  gtk_adjustment_set_value(tracker, value);
+  unsigned int page_id = zathura_document_get_current_page_number(zathura->document);
+  zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
+  if (page == NULL) {
+    return;
+  }
+
+  GtkAdjustment* vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
+  GtkAdjustment* hadj = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
+
+  double position_x = zathura_document_get_position_x(zathura->document);
+  double position_y = zathura_document_get_position_y(zathura->document);
+
+  zathura_adjustment_set_value_from_ratio(vadj, position_y);
+  zathura_adjustment_set_value_from_ratio(hadj, position_x);
+
+  statusbar_page_number_update(zathura);
 }
 
 void
-cb_adjustment_track_bounds(GtkAdjustment* adjustment, gpointer data)
-{
-  GtkAdjustment* tracker = data;
-  gdouble value = gtk_adjustment_get_value(adjustment);
-  gdouble lower = gtk_adjustment_get_lower(adjustment);
-  gdouble upper = gtk_adjustment_get_upper(adjustment);
-  gdouble page_size = gtk_adjustment_get_page_size(adjustment);
-  gtk_adjustment_set_value(tracker, value);
-  gtk_adjustment_set_lower(tracker, lower);
-  gtk_adjustment_set_upper(tracker, upper);
-  gtk_adjustment_set_page_size(tracker, page_size);
-}
-
-void
-cb_pages_per_row_value_changed(girara_session_t* session, const char* UNUSED(name), girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
+cb_page_layout_value_changed(girara_session_t* session, const char* UNUSED(name), girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
 {
   g_return_if_fail(value != NULL);
   g_return_if_fail(session != NULL);
   g_return_if_fail(session->global.data != NULL);
   zathura_t* zathura = session->global.data;
 
-  int pages_per_row = *(int*) value;
-
-  if (pages_per_row < 1) {
-    pages_per_row = 1;
-  }
-
-  unsigned int first_page_column = 1;
-  girara_setting_get(session, "first-page-column", &first_page_column);
-
-  page_widget_set_mode(zathura, pages_per_row, first_page_column);
-
-  if (zathura->document != NULL) {
-    unsigned int current_page = zathura_document_get_current_page_number(zathura->document);
-    page_set_delayed(zathura, current_page);
-  }
-}
-
-void
-cb_first_page_column_value_changed(girara_session_t* session, const char* UNUSED(name), girara_setting_type_t UNUSED(type), void* value, void* UNUSED(data))
-{
-  g_return_if_fail(value != NULL);
-  g_return_if_fail(session != NULL);
-  g_return_if_fail(session->global.data != NULL);
-  zathura_t* zathura = session->global.data;
-
-  int first_page_column = *(int*) value;
-
-  if (first_page_column < 1) {
-    first_page_column = 1;
+  if (zathura->document == NULL) {
+    /* no document has been openend yet */
+    return;
   }
 
   unsigned int pages_per_row = 1;
   girara_setting_get(session, "pages-per-row", &pages_per_row);
 
-  page_widget_set_mode(zathura, pages_per_row, first_page_column);
+  unsigned int first_page_column = 1;
+  girara_setting_get(session, "first-page-column", &first_page_column);
 
-  if (zathura->document != NULL) {
-    unsigned int current_page = zathura_document_get_current_page_number(zathura->document);
-    page_set_delayed(zathura, current_page);
-  }
+  unsigned int page_padding = 1;
+  girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
+
+  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column);
+  zathura_document_set_page_layout(zathura->document, page_padding, pages_per_row, first_page_column);
 }
 
 void
@@ -267,10 +267,7 @@ cb_index_row_activated(GtkTreeView* tree_view, GtkTreePath* path,
     }
 
     sc_toggle_index(zathura->ui.session, NULL, NULL, 0);
-
-    /* zathura_jumplist_save is called when entering index mode */
     zathura_link_evaluate(zathura, index_element->link);
-    zathura_jumplist_add(zathura);
   }
 
   g_object_unref(model);
@@ -326,9 +323,7 @@ handle_link(GtkEntry* entry, girara_session_t* session,
         invalid_index = false;
         switch (action) {
           case ZATHURA_LINK_ACTION_FOLLOW:
-            zathura_jumplist_save(zathura);
             zathura_link_evaluate(zathura, link);
-            zathura_jumplist_add(zathura);
             break;
           case ZATHURA_LINK_ACTION_DISPLAY:
             zathura_link_display(zathura, link);
@@ -359,6 +354,13 @@ cb_sc_display_link(GtkEntry* entry, girara_session_t* session)
   return handle_link(entry, session, ZATHURA_LINK_ACTION_DISPLAY);
 }
 
+static gboolean
+file_monitor_reload(void* data)
+{
+  sc_reload((girara_session_t*) data, NULL, NULL, 0);
+  return FALSE;
+}
+
 void
 cb_file_monitor(GFileMonitor* monitor, GFile* file, GFile* UNUSED(other_file), GFileMonitorEvent event, girara_session_t* session)
 {
@@ -369,9 +371,7 @@ cb_file_monitor(GFileMonitor* monitor, GFile* file, GFile* UNUSED(other_file), G
   switch (event) {
     case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
     case G_FILE_MONITOR_EVENT_CREATED:
-      gdk_threads_enter();
-      sc_reload(session, NULL, NULL, 0);
-      gdk_threads_leave();
+      g_main_context_invoke(NULL, file_monitor_reload, session);
       break;
     default:
       return;
@@ -449,23 +449,19 @@ error_ret:
 }
 
 bool
-cb_view_resized(GtkWidget* UNUSED(widget), GtkAllocation* allocation, zathura_t* zathura)
+cb_view_resized(GtkWidget* UNUSED(widget), GtkAllocation* UNUSED(allocation), zathura_t* zathura)
 {
   if (zathura == NULL || zathura->document == NULL) {
     return false;
   }
 
-  static int height = -1;
-  static int width = -1;
+  /* adjust the scale according to settings. If nothing needs to be resized,
+     it does not trigger the resize event.
 
-  /* adjust only if the allocation changed */
-  if (width != allocation->width || height != allocation->height) {
-    girara_argument_t argument = { zathura_document_get_adjust_mode(zathura->document), NULL };
-    sc_adjust_window(zathura->ui.session, &argument, NULL, 0);
-
-    width  = allocation->width;
-    height = allocation->height;
-  }
+     The right viewport size is already in the document object, due to a
+     previous call to adjustment_changed. We don't want to use the allocation in
+     here, because we would have to subtract scrollbars, etc. */
+  adjust_view(zathura);
 
   return false;
 }
@@ -480,11 +476,11 @@ cb_setting_recolor_change(girara_session_t* session, const char* name,
   g_return_if_fail(name != NULL);
   zathura_t* zathura = session->global.data;
 
-  bool bool_value = *((bool*) value);
+  const bool bool_value = *((bool*) value);
 
-  if (zathura->global.recolor != bool_value) {
-    zathura->global.recolor = bool_value;
-    render_all(zathura);
+  if (zathura->sync.render_thread != NULL && zathura_renderer_recolor_enabled(zathura->sync.render_thread) != bool_value) {
+     zathura_renderer_enable_recolor(zathura->sync.render_thread, bool_value);
+     render_all(zathura);
   }
 }
 
@@ -498,11 +494,11 @@ cb_setting_recolor_keep_hue_change(girara_session_t* session, const char* name,
   g_return_if_fail(name != NULL);
   zathura_t* zathura = session->global.data;
 
-  bool bool_value = *((bool*) value);
+  const bool bool_value = *((bool*) value);
 
-  if (zathura->global.recolor_keep_hue != bool_value) {
-    zathura->global.recolor_keep_hue = bool_value;
-    render_all(zathura);
+  if (zathura->sync.render_thread != NULL && zathura_renderer_recolor_hue_enabled(zathura->sync.render_thread) != bool_value) {
+     zathura_renderer_enable_recolor_hue(zathura->sync.render_thread, bool_value);
+     render_all(zathura);
   }
 }
 
@@ -521,15 +517,59 @@ cb_unknown_command(girara_session_t* session, const char* input)
   }
 
   /* check for number */
-  for (unsigned int i = 0; i < strlen(input); i++) {
+  const size_t size = strlen(input);
+  for (size_t i = 0; i < size; i++) {
     if (g_ascii_isdigit(input[i]) == FALSE) {
       return false;
     }
   }
 
-  zathura_jumplist_save(zathura);
+  zathura_jumplist_add(zathura);
   page_set(zathura, atoi(input) - 1);
   zathura_jumplist_add(zathura);
 
   return true;
 }
+
+void
+cb_page_widget_text_selected(ZathuraPage* page, const char* text, void* data)
+{
+  g_return_if_fail(page != NULL);
+  g_return_if_fail(text != NULL);
+  g_return_if_fail(data != NULL);
+
+  zathura_t* zathura = data;
+  GdkAtom* selection = get_selection(zathura);
+
+  /* copy to clipboard */
+  if (selection != NULL) {
+    gtk_clipboard_set_text(gtk_clipboard_get(*selection), text, -1);
+
+    char* stripped_text = g_strdelimit(g_strdup(text), "\n\t\r\n", ' ');
+    char* escaped_text = g_markup_printf_escaped(
+        _("Copied selected text to clipboard: %s"), stripped_text);
+    g_free(stripped_text);
+
+    girara_notify(zathura->ui.session, GIRARA_INFO, "%s", escaped_text);
+    g_free(escaped_text);
+  }
+
+  g_free(selection);
+}
+
+void
+cb_page_widget_image_selected(ZathuraPage* page, GdkPixbuf* pixbuf, void* data)
+{
+  g_return_if_fail(page != NULL);
+  g_return_if_fail(pixbuf != NULL);
+  g_return_if_fail(data != NULL);
+
+  zathura_t* zathura = data;
+  GdkAtom* selection = get_selection(zathura);
+
+  if (selection != NULL) {
+    gtk_clipboard_set_image(gtk_clipboard_get(*selection), pixbuf);
+  }
+
+  g_free(selection);
+}
