@@ -18,6 +18,7 @@
 #include "plugin.h"
 #include "internal.h"
 #include "render.h"
+#include "adjustment.h"
 
 #include <girara/session.h>
 #include <girara/settings.h>
@@ -44,19 +45,24 @@ cmd_bookmark_create(girara_session_t* session, girara_list_t* argument_list)
 
   const char* bookmark_name = girara_list_nth(argument_list, 0);
   zathura_bookmark_t* bookmark = zathura_bookmark_get(zathura, bookmark_name);
-  if (bookmark != NULL) {
-    bookmark->page = zathura_document_get_current_page_number(zathura->document) + 1;
-    girara_notify(session, GIRARA_INFO, _("Bookmark successfuly updated: %s"), bookmark_name);
-    return true;
-  }
+  bool update = bookmark != NULL ? true : false;
 
   bookmark = zathura_bookmark_add(zathura, bookmark_name, zathura_document_get_current_page_number(zathura->document) + 1);
   if (bookmark == NULL) {
-    girara_notify(session, GIRARA_ERROR, _("Could not create bookmark: %s"), bookmark_name);
+    if (update == true) {
+      girara_notify(session, GIRARA_ERROR, _("Could not update bookmark: %s"), bookmark_name);
+    } else {
+      girara_notify(session, GIRARA_ERROR, _("Could not create bookmark: %s"), bookmark_name);
+    }
     return false;
+  } else {
+    if (update == true) {
+      girara_notify(session, GIRARA_INFO, _("Bookmark successfully updated: %s"), bookmark_name);
+    } else {
+      girara_notify(session, GIRARA_INFO, _("Bookmark successfully created: %s"), bookmark_name);
+    }
   }
 
-  girara_notify(session, GIRARA_INFO, _("Bookmark successfuly created: %s"), bookmark_name);
   return true;
 }
 
@@ -111,7 +117,14 @@ cmd_bookmark_open(girara_session_t* session, girara_list_t* argument_list)
     return false;
   }
 
-  return page_set(zathura, bookmark->page - 1);
+  zathura_jumplist_add(zathura);
+  page_set(zathura, bookmark->page - 1);
+  if (bookmark->x != DBL_MIN && bookmark->y != DBL_MIN) {
+    position_set(zathura, bookmark->x, bookmark->y);
+  }
+  zathura_jumplist_add(zathura);
+
+  return true;
 }
 
 bool
@@ -145,15 +158,15 @@ cmd_info(girara_session_t* session, girara_list_t* UNUSED(argument_list))
     zathura_document_information_type_t field;
   };
 
-  struct meta_field meta_fields[] = {
-    { "Title",            ZATHURA_DOCUMENT_INFORMATION_TITLE },
-    { "Author",           ZATHURA_DOCUMENT_INFORMATION_AUTHOR },
-    { "Subject",          ZATHURA_DOCUMENT_INFORMATION_SUBJECT },
-    { "Keywords",         ZATHURA_DOCUMENT_INFORMATION_KEYWORDS },
-    { "Creator",          ZATHURA_DOCUMENT_INFORMATION_CREATOR },
-    { "Producer",         ZATHURA_DOCUMENT_INFORMATION_PRODUCER },
-    { "Creation date",    ZATHURA_DOCUMENT_INFORMATION_CREATION_DATE },
-    { "Modiciation date", ZATHURA_DOCUMENT_INFORMATION_MODIFICATION_DATE }
+  const struct meta_field meta_fields[] = {
+    { _("Title"),             ZATHURA_DOCUMENT_INFORMATION_TITLE },
+    { _("Author"),            ZATHURA_DOCUMENT_INFORMATION_AUTHOR },
+    { _("Subject"),           ZATHURA_DOCUMENT_INFORMATION_SUBJECT },
+    { _("Keywords"),          ZATHURA_DOCUMENT_INFORMATION_KEYWORDS },
+    { _("Creator"),           ZATHURA_DOCUMENT_INFORMATION_CREATOR },
+    { _("Producer"),          ZATHURA_DOCUMENT_INFORMATION_PRODUCER },
+    { _("Creation date"),     ZATHURA_DOCUMENT_INFORMATION_CREATION_DATE },
+    { _("Modification date"), ZATHURA_DOCUMENT_INFORMATION_MODIFICATION_DATE }
   };
 
   girara_list_t* information = zathura_document_get_information(zathura->document, NULL);
@@ -339,7 +352,6 @@ cmd_search(girara_session_t* session, const char* input, girara_argument_t* argu
     return false;
   }
 
-  bool firsthit = true;
   zathura_error_t error = ZATHURA_ERROR_OK;
 
   /* set search direction */
@@ -352,10 +364,6 @@ cmd_search(girara_session_t* session, const char* input, girara_argument_t* argu
   bool nohlsearch = false;
   girara_setting_get(session, "nohlsearch", &nohlsearch);
 
-  if (nohlsearch == false) {
-    document_draw_search_results(zathura, true);
-  }
-
   /* search pages */
   for (unsigned int page_id = 0; page_id < number_of_pages; ++page_id) {
     unsigned int index = (page_id + current_page_number) % number_of_pages;
@@ -367,9 +375,9 @@ cmd_search(girara_session_t* session, const char* input, girara_argument_t* argu
     GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
     g_object_set(page_widget, "draw-links", FALSE, NULL);
 
-    render_lock(zathura->sync.render_thread);
+    zathura_renderer_lock(zathura->sync.render_thread);
     girara_list_t* result = zathura_page_search_text(page, input, &error);
-    render_unlock(zathura->sync.render_thread);
+    zathura_renderer_unlock(zathura->sync.render_thread);
 
     if (result == NULL || girara_list_size(result) == 0) {
       girara_list_free(result);
@@ -383,20 +391,21 @@ cmd_search(girara_session_t* session, const char* input, girara_argument_t* argu
     }
 
     g_object_set(page_widget, "search-results", result, NULL);
-    if (firsthit == true) {
-      if (page_id != 0) {
-        page_set_delayed(zathura, zathura_page_get_index(page));
-      }
-      if (argument->n == BACKWARD) {
-        /* start at bottom hit in page */
-        g_object_set(page_widget, "search-current", girara_list_size(result) - 1, NULL);
-      } else {
-        g_object_set(page_widget, "search-current", 0, NULL);
-      }
-      firsthit = false;
+
+    if (argument->n == BACKWARD) {
+      /* start at bottom hit in page */
+      g_object_set(page_widget, "search-current", girara_list_size(result) - 1, NULL);
+    } else {
+      g_object_set(page_widget, "search-current", 0, NULL);
     }
   }
 
+  girara_argument_t* arg = g_malloc0(sizeof(girara_argument_t));
+
+  arg->n = FORWARD;
+  sc_search(session, arg, NULL, 0);
+  g_free(arg);
+
   return true;
 }
 
