@@ -19,6 +19,14 @@
 #include "page-widget.h"
 #include "adjustment.h"
 
+#ifndef MIN
+#define MIN(a,b) (((a)<(b))?(a):(b))
+#endif
+
+#ifndef MAX
+#define MAX(a,b) (((a)>(b))?(a):(b))
+#endif
+
 /* Helper function; see sc_display_link and sc_follow. */
 static bool
 draw_links(zathura_t* zathura)
@@ -34,7 +42,7 @@ draw_links(zathura_t* zathura)
     }
 
     GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
-    g_object_set(page_widget, "search-results", NULL, NULL);
+    g_object_set(page_widget, "draw-search-results", FALSE, NULL);
     if (zathura_page_get_visibility(page) == true) {
       g_object_set(page_widget, "draw-links", TRUE, NULL);
 
@@ -71,9 +79,10 @@ sc_abort(girara_session_t* session, girara_argument_t* UNUSED(argument),
         continue;
       }
 
-      g_object_set(zathura_page_get_widget(zathura, page), "draw-links", FALSE, NULL);
+      GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
+      g_object_set(page_widget, "draw-links", FALSE, NULL);
       if (clear_search == true) {
-        g_object_set(zathura_page_get_widget(zathura, page), "search-results", NULL, NULL);
+        g_object_set(page_widget, "draw-search-results", FALSE, NULL);
       }
     }
   }
@@ -93,97 +102,8 @@ sc_adjust_window(girara_session_t* session, girara_argument_t* argument,
   zathura_t* zathura = session->global.data;
   g_return_val_if_fail(argument != NULL, false);
 
-  unsigned int pages_per_row = 1;
-  girara_setting_get(session, "pages-per-row", &pages_per_row);
-
-  unsigned int first_page_column = 1;
-  girara_setting_get(session, "first-page-column", &first_page_column);
-
-  int padding = 1;
-  girara_setting_get(zathura->ui.session, "page-padding", &padding);
-
-  if (zathura->ui.page_widget == NULL || zathura->document == NULL) {
-    goto error_ret;
-  }
-
   zathura_document_set_adjust_mode(zathura->document, argument->n);
-  if (argument->n == ZATHURA_ADJUST_NONE) {
-    /* there is nothing todo */
-    goto error_ret;
-  }
-
-  /* get window size */
-  GtkAllocation allocation;
-  gtk_widget_get_allocation(session->gtk.view, &allocation);
-  unsigned int width  = allocation.width;
-  unsigned int height = allocation.height;
-
-  /* scrollbar spacing */
-  gint spacing;
-  gtk_widget_style_get(session->gtk.view, "scrollbar_spacing", &spacing, NULL);
-  width -= spacing;
-
-  /* correct view size */
-  if (gtk_widget_get_visible(GTK_WIDGET(session->gtk.inputbar)) == true) {
-    gtk_widget_get_allocation(session->gtk.inputbar, &allocation);
-    height += allocation.height;
-  }
-
-  double scale = 1.0;
-  unsigned int cell_height = 0, cell_width = 0;
-  unsigned int document_height = 0, document_width = 0;
-
-  zathura_document_set_scale(zathura->document, scale);
-  zathura_document_get_cell_size(zathura->document, &cell_height, &cell_width);
-  zathura_get_document_size(zathura, cell_height, cell_width,
-                            &document_height, &document_width);
-
-  double page_ratio   = (double)cell_height / (double)document_width;
-  double window_ratio = (double)height / (double)width;
-
-  if (argument->n == ZATHURA_ADJUST_WIDTH ||
-      (argument->n == ZATHURA_ADJUST_BESTFIT && page_ratio < window_ratio)) {
-    scale = (double)(width - (pages_per_row - 1) * padding) /
-            (double)(pages_per_row * cell_width);
-    zathura_document_set_scale(zathura->document, scale);
-
-    bool show_scrollbars = false;
-    girara_setting_get(session, "show-scrollbars", &show_scrollbars);
-
-    if (show_scrollbars) {
-      /* If the document is taller than the view, there's a vertical
-       * scrollbar; we need to substract its width from the view's width. */
-      zathura_get_document_size(zathura, cell_height, cell_width,
-                                &document_height, &document_width);
-      if (height < document_height) {
-        GtkWidget* vscrollbar = gtk_scrolled_window_get_vscrollbar(
-            GTK_SCROLLED_WINDOW(session->gtk.view));
-
-        if (vscrollbar != NULL) {
-          GtkRequisition requisition;
-          gtk_widget_get_requisition(vscrollbar, &requisition);
-          if (0 < requisition.width && (unsigned)requisition.width < width) {
-            width -= requisition.width;
-            scale = (double)(width - (pages_per_row - 1) * padding) /
-                    (double)(pages_per_row * cell_width);
-            zathura_document_set_scale(zathura->document, scale);
-          }
-        }
-      }
-    }
-  }
-  else if (argument->n == ZATHURA_ADJUST_BESTFIT) {
-    scale = (double)height / (double)cell_height;
-    zathura_document_set_scale(zathura->document, scale);
-  }
-  else {
-    goto error_ret;
-  }
-
-  /* re-render all pages */
-  render_all(zathura);
-
-error_ret:
+  adjust_view(zathura);
 
   return false;
 }
@@ -265,16 +185,24 @@ sc_focus_inputbar(girara_session_t* session, girara_argument_t* argument, girara
       g_free(tmp);
     }
 
+    GdkAtom* selection = get_selection(zathura);
+
     /* we save the X clipboard that will be clear by "grab_focus" */
-    gchar* x_clipboard_text = gtk_clipboard_wait_for_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY));
+    gchar* x_clipboard_text = NULL;
+
+    if (selection != NULL) {
+      x_clipboard_text = gtk_clipboard_wait_for_text(gtk_clipboard_get(*selection));
+    }
 
     gtk_editable_set_position(GTK_EDITABLE(session->gtk.inputbar_entry), -1);
 
-    if (x_clipboard_text != NULL) {
+    if (x_clipboard_text != NULL && selection != NULL) {
       /* we reset the X clipboard with saved text */
-      gtk_clipboard_set_text(gtk_clipboard_get(GDK_SELECTION_PRIMARY), x_clipboard_text, -1);
+      gtk_clipboard_set_text(gtk_clipboard_get(*selection), x_clipboard_text, -1);
       g_free(x_clipboard_text);
     }
+
+    g_free(selection);
   }
 
   return true;
@@ -312,7 +240,7 @@ sc_goto(girara_session_t* session, girara_argument_t* argument, girara_event_t*
   g_return_val_if_fail(argument != NULL, false);
   g_return_val_if_fail(zathura->document != NULL, false);
 
-  zathura_jumplist_save(zathura);
+  zathura_jumplist_add(zathura);
   if (t != 0) {
     /* add offset */
     t += zathura_document_get_page_offset(zathura->document);
@@ -324,10 +252,6 @@ sc_goto(girara_session_t* session, girara_argument_t* argument, girara_event_t*
     page_set(zathura, zathura_document_get_number_of_pages(zathura->document) - 1);
   }
 
-  /* adjust horizontal position */
-  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-  cb_view_hadjustment_changed(hadjustment, zathura);
-
   zathura_jumplist_add(zathura);
 
   return false;
@@ -464,10 +388,6 @@ sc_navigate(girara_session_t* session, girara_argument_t* argument,
 
   page_set(zathura, new_page);
 
-  /* adjust horizontal position */
-  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-  cb_view_hadjustment_changed(hadjustment, zathura);
-
   return false;
 }
 
@@ -554,7 +474,7 @@ sc_rotate(girara_session_t* session, girara_argument_t* argument,
   /* render all pages again */
   render_all(zathura);
 
-  page_set_delayed(zathura, page_number);
+  page_set(zathura, page_number);
 
   return false;
 }
@@ -575,18 +495,14 @@ sc_scroll(girara_session_t* session, girara_argument_t* argument,
     t = 1;
   }
 
-  GtkAdjustment* adjustment = NULL;
-  if ( (argument->n == LEFT) || (argument->n == FULL_LEFT) || (argument->n == HALF_LEFT) ||
-       (argument->n == RIGHT) || (argument->n == FULL_RIGHT) || (argument->n == HALF_RIGHT)) {
-    adjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-  } else {
-    adjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-  }
+  unsigned int view_width=0, view_height=0;
+  zathura_document_get_viewport_size(zathura->document, &view_height, &view_width);
 
-  gdouble view_size                  = gtk_adjustment_get_page_size(adjustment);
-  gdouble value                      = gtk_adjustment_get_value(adjustment);
-  gdouble max                        = gtk_adjustment_get_upper(adjustment) - view_size;
-  zathura->global.update_page_number = true;
+  unsigned int cell_width=0, cell_height=0;
+  zathura_document_get_cell_size(zathura->document, &cell_height, &cell_width);
+
+  unsigned int doc_width=0, doc_height=0;
+  zathura_document_get_document_size(zathura->document, &doc_height, &doc_width);
 
   float scroll_step = 40;
   girara_setting_get(session, "scroll-step", &scroll_step);
@@ -606,108 +522,111 @@ sc_scroll(girara_session_t* session, girara_argument_t* argument,
   int padding = 1;
   girara_setting_get(session, "page-padding", &padding);
 
-  gdouble new_value;
+  double pos_x = zathura_document_get_position_x(zathura->document);
+  double pos_y = zathura_document_get_position_y(zathura->document);
+  double page_id = zathura_document_get_current_page_number(zathura->document);
+  double direction = 1.0;
 
+  /* if TOP or BOTTOM, go there and we are done */
+  if (argument->n == TOP) {
+    position_set(zathura, -1, 0);
+    return false;
+  } else if (argument->n == BOTTOM) {
+    position_set(zathura, -1, 1.0);
+    return false;
+  }
+
+  /* compute the direction of scrolling */
+  if ( (argument->n == LEFT) || (argument->n == FULL_LEFT) || (argument->n == HALF_LEFT) ||
+       (argument->n == UP) || (argument->n == FULL_UP) || (argument->n == HALF_UP)) {
+    direction = -1.0;
+  } else {
+    direction = 1.0;
+  }
+
+  double vstep = (double)(cell_height + padding) / (double)doc_height;
+  double hstep = (double)(cell_width + padding) / (double)doc_width;
+
+  /* compute new position */
   switch(argument->n) {
     case FULL_UP:
-    case FULL_LEFT:
-      new_value = value - (1.0 - scroll_full_overlap) * view_size - padding;
-      break;
     case FULL_DOWN:
+      pos_y += direction * (1.0 - scroll_full_overlap) * vstep;
+      break;
+
+    case FULL_LEFT:
     case FULL_RIGHT:
-      new_value = value + (1.0 - scroll_full_overlap) * view_size + padding;
+      pos_x += direction * (1.0 - scroll_full_overlap) * hstep;
       break;
+
     case HALF_UP:
-    case HALF_LEFT:
-      new_value = value - ((view_size + padding) / 2);
-      break;
     case HALF_DOWN:
+      pos_y += direction * 0.5 * vstep;
+      break;
+
+    case HALF_LEFT:
     case HALF_RIGHT:
-      new_value = value + ((view_size + padding) / 2);
-      break;
-    case LEFT:
-      new_value = value - scroll_hstep * t;
+      pos_x += direction * 0.5 * hstep;
       break;
+
     case UP:
-      new_value = value - scroll_step * t;
-      break;
-    case RIGHT:
-      new_value = value + scroll_hstep * t;
-      break;
     case DOWN:
-      new_value = value + scroll_step * t;
+      pos_y += direction * t * scroll_step / (double)doc_height;
       break;
-    case TOP:
-      new_value = 0;
+
+    case LEFT:
+    case RIGHT:
+      pos_x += direction * t * scroll_hstep / (double)doc_width;
       break;
-    case BOTTOM:
-      new_value = max;
-      break;
-    default:
-      new_value = value;
   }
 
-  if (scroll_wrap == true) {
-    if (new_value < 0)
-      new_value = max;
-    else if (new_value > max)
-      new_value = 0;
+  /* handle boundaries */
+  double end_x = 0.5 * (double)view_width / (double)doc_width;
+  double end_y = 0.5 * (double)view_height / (double)doc_height;
+
+  double new_x = scroll_wrap ? 1.0 - end_x : end_x;
+  double new_y = scroll_wrap ? 1.0 - end_y : end_y;
+
+  if (pos_x < end_x) {
+    pos_x = new_x;
+  } else if (pos_x > 1.0 - end_x) {
+    pos_x = 1 - new_x;
   }
 
-  if (scroll_page_aware == true) {
-    int page_offset;
-    double page_size;
+  if (pos_y < end_y) {
+    pos_y = new_y;
+  } else if (pos_y > 1.0 - end_y) {
+    pos_y = 1 - new_y;
+  }
 
-    {
-      unsigned int page_id = zathura_document_get_current_page_number(zathura->document);
-      zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
-      page_offset_t offset;
-      page_calculate_offset(zathura, page, &offset);
+  /* snap to the border if we change page */
+  double dummy;
+  unsigned int new_page_id = position_to_page_number(zathura->document, pos_x, pos_y);
+  if (scroll_page_aware == true && page_id != new_page_id) {
+    switch(argument->n) {
+      case FULL_LEFT:
+      case HALF_LEFT:
+        page_number_to_position(zathura->document, new_page_id, 1.0, 0.0, &pos_x, &dummy);
+        break;
 
-      double scale = zathura_document_get_scale(zathura->document);
+      case FULL_RIGHT:
+      case HALF_RIGHT:
+        page_number_to_position(zathura->document, new_page_id, 0.0, 0.0, &pos_x, &dummy);
+        break;
 
-      if ((argument->n == LEFT) || (argument->n == FULL_LEFT) || (argument->n == HALF_LEFT) ||
-          (argument->n == RIGHT) || (argument->n == FULL_RIGHT) || (argument->n == HALF_RIGHT)) {
-        page_offset = offset.x;
-        page_size = zathura_page_get_width(page) * scale;
-      } else {
-        page_offset = offset.y;
-        page_size = zathura_page_get_height(page) * scale;
-      }
+      case FULL_UP:
+      case HALF_UP:
+        page_number_to_position(zathura->document, new_page_id, 0.0, 1.0, &dummy, &pos_y);
+        break;
 
-      page_offset -= padding / 2;
-      page_size   += padding;
-    }
-
-    if ((argument->n == FULL_DOWN) || (argument->n == HALF_DOWN) ||
-        (argument->n == FULL_RIGHT) || (argument->n == HALF_RIGHT)) {
-      if ((page_offset > value) &&
-          (page_offset < value + view_size)) {
-        new_value = page_offset;
-      } else if ((page_offset <= value) &&
-                 (page_offset + page_size < value + view_size)) {
-        new_value = page_offset + page_size + 1;
-      } else if ((page_offset <= value) &&
-                 (page_offset + page_size < new_value + view_size)) {
-        new_value = page_offset + page_size - view_size + 1;
-      }
-    } else if ((argument->n == FULL_UP) || (argument->n == HALF_UP) ||
-               (argument->n == FULL_LEFT) || (argument->n == HALF_LEFT)) {
-      if ((page_offset + 1 >= value) &&
-          (page_offset < value + view_size)) {
-        new_value = page_offset - view_size;
-      } else if ((page_offset <= value) &&
-                 (page_offset + page_size + 1 < value + view_size)) {
-        new_value = page_offset + page_size - view_size;
-      } else if ((page_offset <= value) &&
-                 (page_offset > new_value)) {
-        new_value = page_offset;
-      }
+      case FULL_DOWN:
+      case HALF_DOWN:
+        page_number_to_position(zathura->document, new_page_id, 0.0, 0.0, &dummy, &pos_y);
+        break;
     }
   }
 
-  zathura_adjustment_set_value(adjustment, new_value);
-
+  position_set(zathura, pos_x, pos_y);
   return false;
 }
 
@@ -722,25 +641,55 @@ sc_jumplist(girara_session_t* session, girara_argument_t* argument,
   g_return_val_if_fail(argument != NULL, false);
   g_return_val_if_fail(zathura->document != NULL, false);
 
+  /* if no jumps in the jumplist */
+  if (zathura->jumplist.size == 0) {
+    return true;
+  }
+
+  double x = zathura_document_get_position_x(zathura->document);
+  double y = zathura_document_get_position_y(zathura->document);
+
   zathura_jump_t* jump = NULL;
+  zathura_jump_t* prev_jump = zathura_jumplist_current(zathura);
+  bool go_to_current = false;
+
+  if (zathura_jumplist_has_next(zathura) == false || zathura_jumplist_has_previous(zathura) == false) {
+    if (x == prev_jump->x && y == prev_jump->y) {
+      go_to_current = false;
+    } else {
+      go_to_current = true;
+    }
+  }
+
   switch(argument->n) {
     case FORWARD:
-      zathura_jumplist_save(zathura);
-      zathura_jumplist_forward(zathura);
-      jump = zathura_jumplist_current(zathura);
+      if (go_to_current == true && zathura_jumplist_has_previous(zathura) == false) {
+        jump = zathura_jumplist_current(zathura);
+      } else {
+        zathura_jumplist_forward(zathura);
+        jump = zathura_jumplist_current(zathura);
+      }
       break;
-
     case BACKWARD:
-      zathura_jumplist_save(zathura);
-      zathura_jumplist_backward(zathura);
-      jump = zathura_jumplist_current(zathura);
+      if (go_to_current == true && zathura_jumplist_has_next(zathura) == false) {
+        jump = zathura_jumplist_current(zathura);
+      } else {
+        zathura_jumplist_backward(zathura);
+        jump = zathura_jumplist_current(zathura);
+      }
       break;
   }
 
+  if (jump == prev_jump) {
+    if ((zathura_jumplist_has_previous(zathura) == false && argument->n == BACKWARD) ||
+        (zathura_jumplist_has_next(zathura) == false && argument->n == FORWARD)) {
+      jump = NULL;
+    }
+  }
+
   if (jump != NULL) {
     page_set(zathura, jump->page);
-    const double s = zathura_document_get_scale(zathura->document);
-    position_set_delayed(zathura, jump->x * s, jump->y * s);
+    position_set(zathura, jump->x, jump->y);
   }
 
   return false;
@@ -757,124 +706,115 @@ sc_bisect(girara_session_t* session, girara_argument_t* argument,
   g_return_val_if_fail(argument != NULL, false);
   g_return_val_if_fail(zathura->document != NULL, false);
 
-  unsigned int number_of_pages, cur_page, prev_page, prev2_page;
-  bool have_prev, have_prev2;
-
-  zathura_jump_t* prev_jump = NULL;
-  zathura_jump_t* prev2_jump = NULL;
-
-  number_of_pages= zathura_document_get_number_of_pages(zathura->document);
-  cur_page = zathura_document_get_current_page_number(zathura->document);
-
-  prev_page = prev2_page = 0;
-  have_prev = have_prev2 = false;
-
-  /* save position at current jump point */
-  zathura_jumplist_save(zathura);
+  const unsigned int num_pages = zathura_document_get_number_of_pages(zathura->document);
+  const unsigned int cur_page = zathura_document_get_current_page_number(zathura->document);
 
   /* process arguments */
   int direction;
-  if (t > 0 && t <= number_of_pages) {
-    /* jump to page t, and bisect between cur_page and t */
-    page_set(zathura, t-1);
-    zathura_jumplist_add(zathura);
-    if (t-1 > cur_page) {
+  if (t > 0 && t <= num_pages) {
+    /* bisect between cur_page and t */
+    t -= 1;
+    if (t == cur_page) {
+      /* nothing to do */
+      return false;
+    }
+    else if (t > cur_page) {
+      zathura->bisect.start = cur_page;
+      zathura->bisect.end = t;
       direction = BACKWARD;
     } else {
+      zathura->bisect.start = t;
+      zathura->bisect.end = cur_page;
       direction = FORWARD;
     }
-
   } else if (argument != NULL)  {
     direction = argument->n;
 
+    /* setup initial bisect range */
+    zathura_jump_t* jump = zathura_jumplist_current(zathura);
+    if (jump == NULL) {
+      girara_debug("bisecting between first and last page because there are no jumps");
+      zathura->bisect.start = 0;
+      zathura->bisect.end = num_pages - 1;
+
+    } else if (jump->page != cur_page || jump->page != zathura->bisect.last_jump) {
+      girara_debug("last jump doesn't match up, starting new bisecting");
+      zathura->bisect.start = 0;
+      zathura->bisect.end = num_pages - 1;
+
+      unsigned int prev_page;
+      if (direction == FORWARD) {
+        prev_page = num_pages - 1;
+      } else {
+        prev_page = 0;
+      }
+
+      /* check if we have previous jumps */
+      if (zathura_jumplist_has_previous(zathura) == true) {
+        zathura_jumplist_backward(zathura);
+        jump = zathura_jumplist_current(zathura);
+        if (jump != NULL) {
+          prev_page = jump->page;
+        }
+        zathura_jumplist_forward(zathura);
+      }
+
+      zathura->bisect.start = MIN(prev_page, cur_page);
+      zathura->bisect.end = MAX(prev_page, cur_page);
+      zathura->bisect.last_jump = cur_page;
+    }
   } else {
     return false;
   }
 
-  cur_page = zathura_document_get_current_page_number(zathura->document);
-
-  if (zathura_jumplist_has_previous(zathura)) {
-    /* If there is a previous jump, get its page */
-    zathura_jumplist_backward(zathura);
-    prev_jump = zathura_jumplist_current(zathura);
-    if (prev_jump) {
-      prev_page = prev_jump->page;
-      have_prev = true;
-    }
-
-    if (zathura_jumplist_has_previous(zathura)) {
-      /* If there is a second previous jump, get its page. */
-      zathura_jumplist_backward(zathura);
-      prev2_jump = zathura_jumplist_current(zathura);
-      if (prev2_jump) {
-        prev2_page = prev2_jump->page;
-        have_prev2 = true;
-      }
-      zathura_jumplist_forward(zathura);
-    }
-    zathura_jumplist_forward(zathura);
+  girara_debug("bisecting between %d and %d, at %d", zathura->bisect.start, zathura->bisect.end, cur_page);
+  if (zathura->bisect.start == zathura->bisect.end) {
+    /* nothing to do */
+    return false;
   }
 
-  /* now, we are back at the initial jump. prev_page and prev2_page contain
-   the pages for previous and second previous jump if they exist. */
+  unsigned int next_page = cur_page;
+  unsigned int next_start = zathura->bisect.start;
+  unsigned int next_end = zathura->bisect.end;
 
-  /* bisect */
+  /* here we have next_start <= next_page <= next_end */
+
+  /* bisect step */
   switch(direction) {
     case FORWARD:
-      if (have_prev && cur_page <= prev_page) {
-        /* add a new jump point */
-        if (cur_page < prev_page) {
-          page_set(zathura, (cur_page + prev_page)/2);
-          zathura_jumplist_add(zathura);
+      if (cur_page != zathura->bisect.end) {
+        next_page = (cur_page + zathura->bisect.end) / 2;
+        if (next_page == cur_page) {
+          ++next_page;
         }
-
-      } else if (have_prev2 && cur_page <= prev2_page) {
-        /* save current position at previous jump point */
-        if (cur_page < prev2_page) {
-          zathura_jumplist_backward(zathura);
-          zathura_jumplist_save(zathura);
-          zathura_jumplist_forward(zathura);
-
-          page_set(zathura, (cur_page + prev2_page)/2);
-          zathura_jumplist_save(zathura);
-        }
-      } else {
-        /* none of prev_page or prev2_page comes after cur_page */
-        page_set(zathura, (cur_page + number_of_pages - 1)/2);
-        zathura_jumplist_add(zathura);
+        next_start = cur_page;
       }
       break;
 
     case BACKWARD:
-      if (have_prev && prev_page <= cur_page) {
-        /* add a new jump point */
-        if (prev_page < cur_page) {
-          page_set(zathura, (cur_page + prev_page)/2);
-          zathura_jumplist_add(zathura);
+      if (cur_page != zathura->bisect.start) {
+        next_page = (cur_page + zathura->bisect.start) / 2;
+        if (next_page == cur_page) {
+          --next_page;
         }
-
-      } else if (have_prev2 && prev2_page <= cur_page) {
-        /* save current position at previous jump point */
-        if (prev2_page < cur_page) {
-          zathura_jumplist_backward(zathura);
-          zathura_jumplist_save(zathura);
-          zathura_jumplist_forward(zathura);
-
-          page_set(zathura, (cur_page + prev2_page)/2);
-          zathura_jumplist_save(zathura);
-        }
-
-      } else {
-        /* none of prev_page or prev2_page comes before cur_page */
-        page_set(zathura, cur_page/2);
-        zathura_jumplist_add(zathura);
+        next_end = cur_page;
       }
       break;
   }
 
-  /* adjust horizontal position */
-  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-  cb_view_hadjustment_changed(hadjustment, zathura);
+  if (next_page == cur_page) {
+    /* nothing to do */
+    return false;
+  }
+
+  girara_debug("bisecting between %d and %d, jumping to %d", zathura->bisect.start, zathura->bisect.end, next_page);
+  zathura->bisect.last_jump = next_page;
+  zathura->bisect.start = next_start;
+  zathura->bisect.end = next_end;
+
+  zathura_jumplist_add(zathura);
+  page_set(zathura, next_page);
+  zathura_jumplist_add(zathura);
 
   return false;
 }
@@ -890,8 +830,23 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
   g_return_val_if_fail(argument != NULL, false);
   g_return_val_if_fail(zathura->document != NULL, false);
 
-  const int num_pages = zathura_document_get_number_of_pages(zathura->document);
-  const int cur_page  = zathura_document_get_current_page_number(zathura->document);
+  const unsigned int num_pages = zathura_document_get_number_of_pages(zathura->document);
+  const unsigned int cur_page  = zathura_document_get_current_page_number(zathura->document);
+  GtkWidget *cur_page_widget = zathura_page_get_widget(zathura, zathura_document_get_page(zathura->document, cur_page));
+  bool nohlsearch, first_time_after_abort, draw;
+
+  nohlsearch = first_time_after_abort = draw = false;
+  girara_setting_get(session, "nohlsearch", &nohlsearch);
+
+  if (nohlsearch == false) {
+    g_object_get(cur_page_widget, "draw-search-results", &draw, NULL);
+
+    if (draw == false) {
+      first_time_after_abort = true;
+    }
+
+    document_draw_search_results(zathura, true);
+  }
 
   int diff = argument->n == FORWARD ? 1 : -1;
   if (zathura->global.search_direction == BACKWARD)
@@ -900,7 +855,7 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
   zathura_page_t* target_page = NULL;
   int target_idx = 0;
 
-  for (int page_id = 0; page_id < num_pages; ++page_id) {
+  for (unsigned int page_id = 0; page_id < num_pages; ++page_id) {
     int tmp = cur_page + diff * page_id;
     zathura_page_t* page = zathura_document_get_page(zathura->document, (tmp + num_pages) % num_pages);
     if (page == NULL) {
@@ -910,12 +865,17 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
     GtkWidget* page_widget = zathura_page_get_widget(zathura, page);
 
     int num_search_results = 0, current = -1;
-    g_object_get(page_widget, "search-current", &current,
-                 "search-length", &num_search_results, NULL);
+    g_object_get(page_widget, "search-current", &current, "search-length", &num_search_results, NULL);
     if (num_search_results == 0 || current == -1) {
       continue;
     }
 
+    if (first_time_after_abort == true || (tmp + num_pages) % num_pages != cur_page) {
+      target_page = page;
+      target_idx = diff == 1 ? 0 : num_search_results - 1;
+      break;
+    }
+
     if (diff == 1 && current < num_search_results - 1) {
       /* the next result is on the same page */
       target_page = page;
@@ -925,14 +885,11 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
       target_idx = current - 1;
     } else {
       /* the next result is on a different page */
-      zathura_jumplist_save(zathura);
-
       g_object_set(page_widget, "search-current", -1, NULL);
 
       for (int npage_id = 1; page_id < num_pages; ++npage_id) {
         int ntmp = cur_page + diff * (page_id + npage_id);
         zathura_page_t* npage = zathura_document_get_page(zathura->document, (ntmp + 2*num_pages) % num_pages);
-        zathura_document_set_current_page_number(zathura->document, zathura_page_get_index(npage));
         GtkWidget* npage_page_widget = zathura_page_get_widget(zathura, npage);
         g_object_get(npage_page_widget, "search-length", &num_search_results, NULL);
         if (num_search_results != 0) {
@@ -941,8 +898,6 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
           break;
         }
       }
-
-      zathura_jumplist_add(zathura);
     }
 
     break;
@@ -956,20 +911,36 @@ sc_search(girara_session_t* session, girara_argument_t* argument,
 
     zathura_rectangle_t* rect = girara_list_nth(results, target_idx);
     zathura_rectangle_t rectangle = recalc_rectangle(target_page, *rect);
-    page_offset_t offset;
-    page_calculate_offset(zathura, target_page, &offset);
-
-    GtkAdjustment* view_vadjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-    int y = offset.y - gtk_adjustment_get_page_size(view_vadjustment) / 2 + rectangle.y1;
-    zathura_adjustment_set_value(view_vadjustment, y);
 
     bool search_hadjust = true;
     girara_setting_get(session, "search-hadjust", &search_hadjust);
+
+    /* compute the position of the center of the page */
+    double pos_x = 0;
+    double pos_y = 0;
+    page_number_to_position(zathura->document, zathura_page_get_index(target_page),
+                            0.5, 0.5, &pos_x, &pos_y);
+
+    /* correction to center the current result                          */
+    /* NOTE: rectangle is in viewport units, already scaled and rotated */
+    unsigned int cell_height = 0;
+    unsigned int cell_width = 0;
+    zathura_document_get_cell_size(zathura->document, &cell_height, &cell_width);
+
+    unsigned int doc_height = 0;
+    unsigned int doc_width = 0;
+    zathura_document_get_document_size(zathura->document, &doc_height, &doc_width);
+
+    pos_y += (rectangle.y1 - (double)cell_height/2) / (double)doc_height;
+
     if (search_hadjust == true) {
-      GtkAdjustment* view_hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-      int x = offset.x - gtk_adjustment_get_page_size(view_hadjustment) / 2 + rectangle.x1;
-      zathura_adjustment_set_value(view_hadjustment, x);
+      pos_x += (rectangle.x1 - (double)cell_width/2) / (double)doc_width;
     }
+
+    /* move to position */
+    zathura_jumplist_add(zathura);
+    position_set(zathura, pos_x, pos_y);
+    zathura_jumplist_add(zathura);
   }
 
   return false;
@@ -1146,26 +1117,16 @@ sc_toggle_index(girara_session_t* session, girara_argument_t* UNUSED(argument),
     gtk_widget_show(treeview);
   }
 
-  static double vvalue = 0;
-  static double hvalue = 0;
-
   if (gtk_widget_get_visible(GTK_WIDGET(zathura->ui.index))) {
     girara_set_view(session, zathura->ui.page_widget_alignment);
     gtk_widget_hide(GTK_WIDGET(zathura->ui.index));
     girara_mode_set(zathura->ui.session, zathura->modes.normal);
 
-    /* reset adjustment */
-    position_set_delayed(zathura, hvalue, vvalue);
+    /* refresh view */
+    refresh_view(zathura);
   } else {
-    /* save adjustment */
-    GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-    GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(session->gtk.view));
-
-    vvalue = gtk_adjustment_get_value(vadjustment);
-    hvalue = gtk_adjustment_get_value(hadjustment);
-
     /* save current position to the jumplist */
-    zathura_jumplist_save(zathura);
+    zathura_jumplist_add(zathura);
 
     girara_set_view(session, zathura->ui.index);
     gtk_widget_show(GTK_WIDGET(zathura->ui.index));
@@ -1253,7 +1214,7 @@ sc_toggle_fullscreen(girara_session_t* session, girara_argument_t*
     /* reset scale */
     zathura_document_set_scale(zathura->document, zoom);
     render_all(zathura);
-    page_set_delayed(zathura, zathura_document_get_current_page_number(zathura->document));
+    refresh_view(zathura);
 
     /* setm ode */
     girara_mode_set(session, zathura->modes.normal);
@@ -1281,7 +1242,7 @@ sc_toggle_fullscreen(girara_session_t* session, girara_argument_t*
 
     /* set full screen */
     gtk_window_fullscreen(GTK_WINDOW(session->gtk.window));
-    page_set_delayed(zathura, zathura_document_get_current_page_number(zathura->document));
+    refresh_view(zathura);
 
     /* setm ode */
     girara_mode_set(session, zathura->modes.fullscreen);
@@ -1358,6 +1319,7 @@ sc_zoom(girara_session_t* session, girara_argument_t* argument, girara_event_t*
   }
 
   render_all(zathura);
+  refresh_view(zathura);
 
   return false;
 }
