@@ -86,35 +86,6 @@ zathura_rectangle_t rotate_rectangle(zathura_rectangle_t rectangle, unsigned int
  */
 zathura_rectangle_t recalc_rectangle(zathura_page_t* page, zathura_rectangle_t rectangle);
 
-/**
- * Calculate the page size according to the corrent scaling and rotation if
- * desired.
- * @param page the page
- * @param page_height the resulting page height
- * @param page_width the resultung page width
- * @param rotate honor page's rotation
- * @return real scale after rounding
- */
-double
-page_calc_height_width(zathura_page_t* page, unsigned int* page_height, unsigned int* page_width, bool rotate);
-
-/**
- * Compute the size of the entire document to be displayed (in pixels), taking
- * into account the scale, the layout of the pages, and the padding between
- * them. It should be equal to the allocation of zathura->ui.page_widget once
- * it's shown.
- *
- * @param[in]  zathura                The zathura instance
- * @param[in]  cell_height,cell_width The height and width of a cell containing
- *                                    a single page; it should be obtained
- *                                    using zathura_document_get_cell_size()
- *                                    with the document scale set to 1.0
- * @param[out] height,width           The height and width of the document
- */
-void zathura_get_document_size(zathura_t* zathura,
-                               unsigned int cell_height, unsigned int cell_width,
-                               unsigned int* height, unsigned int* width);
-
 /**
  * Returns the page widget of the page
  *
@@ -138,7 +109,7 @@ void document_draw_search_results(zathura_t* zathura, bool value);
  *
  * @param zathura The zathura instance
  * @param markup Enable markup
- * @return Version string 
+ * @return Version string
  */
 char* zathura_get_version_string(zathura_t* zathura, bool markup);
 
@@ -154,4 +125,14 @@ char* zathura_get_version_string(zathura_t* zathura, bool markup);
  */
 char* replace_substring(const char* string, const char* old, const char* new);
 
+/**
+ * Get a pointer to the GdkAtom of the current clipboard.
+ *
+ * @param zathura The zathura instance
+ *
+ * @return A pointer to a GdkAtom object correspoinding to the current
+ * clipboard, or NULL.
+ */
+GdkAtom* get_selection(zathura_t* zathura);
+
 #endif // UTILS_H
