@@ -13,6 +13,7 @@
 #include <girara/session.h>
 #include <girara/statusbar.h>
 #include <girara/settings.h>
+#include <girara/shortcuts.h>
 #include <glib/gstdio.h>
 #include <glib/gi18n.h>
 
@@ -41,22 +42,11 @@ typedef struct zathura_document_info_s {
   int page_number;
 } zathura_document_info_t;
 
-typedef struct page_set_delayed_s {
-  zathura_t* zathura;
-  unsigned int page;
-} page_set_delayed_t;
-
-typedef struct position_set_delayed_s {
-  zathura_t* zathura;
-  double position_x;
-  double position_y;
-} position_set_delayed_t;
 
 static gboolean document_info_open(gpointer data);
-static bool zathura_page_cache_is_cached(zathura_t* zathura, unsigned int page_index);
-static ssize_t zathura_page_cache_lru_invalidate(zathura_t* zathura);
-static void zathura_page_cache_invalidate_all(zathura_t* zathura);
-static bool zathura_page_cache_is_full(zathura_t* zathura, bool* result);
+static void zathura_jumplist_reset_current(zathura_t* zathura);
+static void zathura_jumplist_append_jump(zathura_t* zathura);
+static void zathura_jumplist_save(zathura_t* zathura);
 
 /* function implementation */
 zathura_t*
@@ -65,8 +55,6 @@ zathura_create(void)
   zathura_t* zathura = g_malloc0(sizeof(zathura_t));
 
   /* global settings */
-  zathura->global.recolor            = false;
-  zathura->global.update_page_number = true;
   zathura->global.search_direction = FORWARD;
 
   /* plugins */
@@ -142,6 +130,21 @@ zathura_init(zathura_t* zathura)
   zathura->ui.session->events.buffer_changed  = cb_buffer_changed;
   zathura->ui.session->events.unknown_command = cb_unknown_command;
 
+  /* zathura signals */
+  zathura->signals.refresh_view = g_signal_new("refresh-view",
+                                               GTK_TYPE_WIDGET,
+                                               G_SIGNAL_RUN_LAST,
+                                               0,
+                                               NULL,
+                                               NULL,
+                                               g_cclosure_marshal_generic,
+                                               G_TYPE_NONE,
+                                               1,
+                                               G_TYPE_POINTER);
+
+  g_signal_connect(G_OBJECT(zathura->ui.session->gtk.view), "refresh-view",
+                   G_CALLBACK(cb_refresh_view), zathura);
+
   /* page view */
 #if (GTK_MAJOR_VERSION == 3)
   zathura->ui.page_widget = gtk_grid_new();
@@ -156,37 +159,23 @@ zathura_init(zathura_t* zathura)
 
   g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "size-allocate", G_CALLBACK(cb_view_resized), zathura);
 
-  /* Setup hadjustment tracker */
   GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(
-      GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-  zathura->ui.hadjustment = zathura_adjustment_clone(hadjustment);
-  g_object_ref_sink(zathura->ui.hadjustment);
+                 GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
 
   /* Connect hadjustment signals */
   g_signal_connect(G_OBJECT(hadjustment), "value-changed",
-      G_CALLBACK(cb_view_vadjustment_value_changed), zathura);
-  g_signal_connect(G_OBJECT(hadjustment), "value-changed",
-      G_CALLBACK(cb_adjustment_track_value), zathura->ui.hadjustment);
+      G_CALLBACK(cb_view_hadjustment_value_changed), zathura);
   g_signal_connect(G_OBJECT(hadjustment), "changed",
       G_CALLBACK(cb_view_hadjustment_changed), zathura);
-  g_signal_connect(G_OBJECT(hadjustment), "changed",
-      G_CALLBACK(cb_adjustment_track_bounds), zathura->ui.hadjustment);
 
-  /* Setup vadjustment tracker */
   GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(
-      GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-  zathura->ui.vadjustment = zathura_adjustment_clone(vadjustment);
-  g_object_ref_sink(zathura->ui.vadjustment);
+                 GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
 
   /* Connect vadjustment signals */
   g_signal_connect(G_OBJECT(vadjustment), "value-changed",
       G_CALLBACK(cb_view_vadjustment_value_changed), zathura);
-  g_signal_connect(G_OBJECT(vadjustment), "value-changed",
-      G_CALLBACK(cb_adjustment_track_value), zathura->ui.vadjustment);
   g_signal_connect(G_OBJECT(vadjustment), "changed",
       G_CALLBACK(cb_view_vadjustment_changed), zathura);
-  g_signal_connect(G_OBJECT(vadjustment), "changed",
-      G_CALLBACK(cb_adjustment_track_bounds), zathura->ui.vadjustment);
 
   /* page view alignment */
   zathura->ui.page_widget_alignment = gtk_alignment_new(0.5, 0.5, 0, 0);
@@ -226,18 +215,6 @@ zathura_init(zathura_t* zathura)
   /* signals */
   g_signal_connect(G_OBJECT(zathura->ui.session->gtk.window), "destroy", G_CALLBACK(cb_destroy), zathura);
 
-  /* set page padding */
-  int page_padding = 1;
-  girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
-
-#if (GTK_MAJOR_VERSION == 3)
-  gtk_grid_set_row_spacing(GTK_GRID(zathura->ui.page_widget), page_padding);
-  gtk_grid_set_column_spacing(GTK_GRID(zathura->ui.page_widget), page_padding);
-#else
-  gtk_table_set_row_spacings(GTK_TABLE(zathura->ui.page_widget), page_padding);
-  gtk_table_set_col_spacings(GTK_TABLE(zathura->ui.page_widget), page_padding);
-#endif
-
   /* database */
   char* database = NULL;
   girara_setting_get(zathura->ui.session, "database", &database);
@@ -269,28 +246,13 @@ zathura_init(zathura_t* zathura)
 
   /* jumplist */
 
-  zathura->jumplist.max_size = 20;
-  girara_setting_get(zathura->ui.session, "jumplist-size", &(zathura->jumplist.max_size));
+  int jumplist_size = 20;
+  girara_setting_get(zathura->ui.session, "jumplist-size", &jumplist_size);
 
-  zathura->jumplist.list = girara_list_new2(g_free);
+  zathura->jumplist.max_size = jumplist_size < 0 ? 0 : jumplist_size;
+  zathura->jumplist.list = NULL;
   zathura->jumplist.size = 0;
   zathura->jumplist.cur = NULL;
-  zathura_jumplist_append_jump(zathura);
-  zathura->jumplist.cur = girara_list_iterator(zathura->jumplist.list);
-
-  /* page cache */
-
-  int cache_size = 0;
-  girara_setting_get(zathura->ui.session, "page-cache-size", &cache_size);
-  if (cache_size <= 0) {
-    girara_warning("page-cache-size is not positive, using %d instead", ZATHURA_PAGE_CACHE_DEFAULT_SIZE);
-    zathura->page_cache.size = ZATHURA_PAGE_CACHE_DEFAULT_SIZE;
-  } else {
-    zathura->page_cache.size = cache_size;
-  }
-
-  zathura->page_cache.cache = g_malloc(zathura->page_cache.size * sizeof(int));
-  zathura_page_cache_invalidate_all(zathura);
 
   return true;
 
@@ -320,13 +282,6 @@ zathura_free(zathura_t* zathura)
     girara_session_destroy(zathura->ui.session);
   }
 
-  if (zathura->ui.hadjustment != NULL) {
-    g_object_unref(G_OBJECT(zathura->ui.hadjustment));
-  }
-  if (zathura->ui.vadjustment != NULL) {
-    g_object_unref(G_OBJECT(zathura->ui.vadjustment));
-  }
-
   /* stdin support */
   if (zathura->stdin_support.file != NULL) {
     g_unlink(zathura->stdin_support.file);
@@ -366,8 +321,6 @@ zathura_free(zathura_t* zathura)
     girara_list_iterator_free(zathura->jumplist.cur);
   }
 
-  g_free(zathura->page_cache.cache);
-
   g_free(zathura);
 }
 
@@ -530,7 +483,7 @@ document_info_open(gpointer data)
       file = prepare_document_open_from_stdin(document_info->zathura);
       if (file == NULL) {
         girara_notify(document_info->zathura->ui.session, GIRARA_ERROR,
-                      "Could not read file from stdin and write it to a temporary file.");
+                      _("Could not read file from stdin and write it to a temporary file."));
       } else {
         document_info->zathura->stdin_support.file = g_strdup(file);
       }
@@ -578,6 +531,9 @@ document_open(zathura_t* zathura, const char* path, const char* password,
       }
       goto error_out;
     }
+    if (error == ZATHURA_ERROR_OK ) {
+      girara_notify(zathura->ui.session, GIRARA_ERROR, _("Unsupported file type. Please install the necessary plugin."));
+    }
     goto error_out;
   }
 
@@ -648,15 +604,18 @@ document_open(zathura_t* zathura, const char* path, const char* password,
     zathura_document_set_adjust_mode(document, ZATHURA_ADJUST_NONE);
   }
 
+  /* initialize bisect state */
+  zathura->bisect.start = 0;
+  zathura->bisect.last_jump = zathura_document_get_current_page_number(document);
+  zathura->bisect.end = number_of_pages - 1;
+
   /* update statusbar */
   bool basename_only = false;
   girara_setting_get(zathura->ui.session, "statusbar-basename", &basename_only);
   if (basename_only == false) {
     girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, file_path);
   } else {
-    char* tmp = g_path_get_basename(file_path);
-    girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, tmp);
-    g_free(tmp);
+    girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.file, zathura_document_get_basename(document));
   }
 
   /* install file monitor */
@@ -703,6 +662,49 @@ document_open(zathura_t* zathura, const char* path, const char* password,
 
   zathura->document = document;
 
+  /* page cache size */
+  int cache_size = 0;
+  girara_setting_get(zathura->ui.session, "page-cache-size", &cache_size);
+  if (cache_size <= 0) {
+    girara_warning("page-cache-size is not positive, using %d instead",
+        ZATHURA_PAGE_CACHE_DEFAULT_SIZE);
+    cache_size = ZATHURA_PAGE_CACHE_DEFAULT_SIZE;
+  }
+
+  /* threads */
+  zathura->sync.render_thread = zathura_renderer_new(cache_size);
+
+  if (zathura->sync.render_thread == NULL) {
+    goto error_free;
+  }
+
+  /* set up recolor info in ZathuraRenderer */
+  char* recolor_dark = NULL;
+  char* recolor_light = NULL;
+  girara_setting_get(zathura->ui.session, "recolor-darkcolor", &recolor_dark);
+  girara_setting_get(zathura->ui.session, "recolor-lightcolor", &recolor_light);
+  zathura_renderer_set_recolor_colors_str(zathura->sync.render_thread,
+      recolor_light, recolor_dark);
+  g_free(recolor_dark);
+  g_free(recolor_light);
+
+  bool recolor = false;
+  girara_setting_get(zathura->ui.session, "recolor", &recolor);
+  zathura_renderer_enable_recolor(zathura->sync.render_thread, recolor);
+  girara_setting_get(zathura->ui.session, "recolor-keephue", &recolor);
+  zathura_renderer_enable_recolor_hue(zathura->sync.render_thread, recolor);
+
+  /* get view port size */
+  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(
+                 GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
+  GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(
+                 GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
+
+  const unsigned int view_width = (unsigned int)floor(gtk_adjustment_get_page_size(hadjustment));
+  zathura_document_set_viewport_width(zathura->document, view_width);
+  const unsigned int view_height = (unsigned int)floor(gtk_adjustment_get_page_size(vadjustment));
+  zathura_document_set_viewport_height(zathura->document, view_height);
+
   /* create blank pages */
   zathura->pages = calloc(number_of_pages, sizeof(GtkWidget*));
   if (zathura->pages == NULL) {
@@ -722,17 +724,19 @@ document_open(zathura_t* zathura, const char* path, const char* password,
 
     zathura->pages[page_id] = page_widget;
 
-    /* set widget size */
-    unsigned int page_height = 0;
-    unsigned int page_width  = 0;
-    page_calc_height_width(page, &page_height, &page_width, true);
-
-    gtk_widget_set_size_request(page_widget, page_width, page_height);
+    g_signal_connect(G_OBJECT(page_widget), "text-selected",
+        G_CALLBACK(cb_page_widget_text_selected), zathura);
+    g_signal_connect(G_OBJECT(page_widget), "image-selected",
+        G_CALLBACK(cb_page_widget_image_selected), zathura);
   }
 
   /* view mode */
-  int pages_per_row = 1;
-  int first_page_column = 1;
+  unsigned int pages_per_row = 1;
+  unsigned int first_page_column = 1;
+  unsigned int page_padding = 1;
+
+  girara_setting_get(zathura->ui.session, "page-padding", &page_padding);
+
   if (file_info.pages_per_row > 0) {
     pages_per_row = file_info.pages_per_row;
   } else {
@@ -747,51 +751,57 @@ document_open(zathura_t* zathura, const char* path, const char* password,
 
   girara_setting_set(zathura->ui.session, "pages-per-row", &pages_per_row);
   girara_setting_set(zathura->ui.session, "first-page-column", &first_page_column);
-  page_widget_set_mode(zathura, pages_per_row, first_page_column);
+
+  page_widget_set_mode(zathura, page_padding, pages_per_row, first_page_column);
+  zathura_document_set_page_layout(zathura->document, page_padding, pages_per_row, first_page_column);
 
   girara_set_view(zathura->ui.session, zathura->ui.page_widget_alignment);
 
-  /* threads */
-  zathura->sync.render_thread = render_init(zathura);
-
-  if (zathura->sync.render_thread == NULL) {
-    goto error_free;
-  }
-
-  for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
-    gtk_widget_realize(zathura->pages[page_id]);
-  }
 
   /* bookmarks */
   zathura_bookmarks_load(zathura, file_path);
 
+  /* jumplist */
+  if (zathura_jumplist_load(zathura, file_path) == false) {
+    zathura->jumplist.list = girara_list_new2(g_free);
+  }
+
   /* update title */
   basename_only = false;
   girara_setting_get(zathura->ui.session, "window-title-basename", &basename_only);
   if (basename_only == false) {
     girara_set_window_title(zathura->ui.session, file_path);
   } else {
-    char* tmp = g_path_get_basename(file_path);
-    girara_set_window_title(zathura->ui.session, tmp);
-    g_free(tmp);
+    girara_set_window_title(zathura->ui.session, zathura_document_get_basename(document));
   }
 
   g_free(file_uri);
 
-  /* adjust window */
-  girara_argument_t argument = { zathura_document_get_adjust_mode(document), NULL };
-  sc_adjust_window(zathura->ui.session, &argument, NULL, 0);
+  /* adjust_view */
+  adjust_view(zathura);
+  for (unsigned int page_id = 0; page_id < number_of_pages; page_id++) {
+    /* set widget size */
+    zathura_page_t* page = zathura_document_get_page(document, page_id);
+    unsigned int page_height = 0;
+    unsigned int page_width  = 0;
 
-  /* set position */
-  if (file_info.position_x != 0 || file_info.position_y != 0) {
-    position_set_delayed(zathura, file_info.position_x, file_info.position_y);
-  } else {
-    page_set_delayed(zathura, zathura_document_get_current_page_number(document));
-    cb_view_vadjustment_value_changed(NULL, zathura);
+    /* adjust_view calls render_all in some cases and render_all calls
+     * gtk_widget_set_size_request. To be sure that it's really called, do it
+     * here once again. */
+    double height = zathura_page_get_height(page);
+    double width = zathura_page_get_width(page);
+    page_calc_height_width(zathura->document, height, width, &page_height, &page_width, true);
+    gtk_widget_set_size_request(zathura->pages[page_id], page_width, page_height);
+
+    /* show widget */
+    gtk_widget_show(zathura->pages[page_id]);
   }
 
-  /* Invalidate all current entries in the page cache */
-  zathura_page_cache_invalidate_all(zathura);
+  /* set position */
+  page_set(zathura, zathura_document_get_current_page_number(document));
+  if (file_info.position_x != 0 || file_info.position_y != 0) {
+    position_set(zathura, file_info.position_x, file_info.position_y);
+  }
 
   return true;
 
@@ -872,6 +882,9 @@ document_close(zathura_t* zathura, bool keep_monitor)
     return false;
   }
 
+  /* stop rendering */
+  zathura_renderer_stop(zathura->sync.render_thread);
+
   /* remove monitor */
   if (keep_monitor == false) {
     if (zathura->file_monitor.monitor != NULL) {
@@ -915,18 +928,22 @@ document_close(zathura_t* zathura, bool keep_monitor)
   girara_setting_get(zathura->ui.session, "first-page-column", &(file_info.first_page_column));
 
   /* get position */
-  GtkScrolledWindow *window = GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view);
-  GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(window);
-  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(window);
-
-  file_info.position_x = gtk_adjustment_get_value(hadjustment);
-  file_info.position_y = gtk_adjustment_get_value(vadjustment);
+  file_info.position_x = zathura_document_get_position_x(zathura->document);
+  file_info.position_y = zathura_document_get_position_y(zathura->document);
 
   /* save file info */
   zathura_db_set_fileinfo(zathura->database, path, &file_info);
 
+  /* save jumplist */
+  zathura_db_save_jumplist(zathura->database, path, zathura->jumplist.list);
+  girara_list_iterator_free(zathura->jumplist.cur);
+  zathura->jumplist.cur = NULL;
+  girara_list_free(zathura->jumplist.list);
+  zathura->jumplist.list = NULL;
+  zathura->jumplist.size = 0;
+
   /* release render thread */
-  render_free(zathura->sync.render_thread);
+  g_object_unref(zathura->sync.render_thread);
   zathura->sync.render_thread = NULL;
 
   /* remove widgets */
@@ -962,31 +979,6 @@ document_close(zathura_t* zathura, bool keep_monitor)
   return true;
 }
 
-static gboolean
-page_set_delayed_impl(gpointer data)
-{
-  page_set_delayed_t* p = data;
-  page_set(p->zathura, p->page);
-
-  g_free(p);
-  return FALSE;
-}
-
-bool
-page_set_delayed(zathura_t* zathura, unsigned int page_id)
-{
-  if (zathura == NULL || zathura->document == NULL ||
-      (page_id >= zathura_document_get_number_of_pages(zathura->document))) {
-    return false;
-  }
-
-  page_set_delayed_t* p = g_malloc(sizeof(page_set_delayed_t));
-  p->zathura = zathura;
-  p->page = page_id;
-  gdk_threads_add_idle(page_set_delayed_impl, p);
-  return true;
-}
-
 bool
 page_set(zathura_t* zathura, unsigned int page_id)
 {
@@ -994,30 +986,17 @@ page_set(zathura_t* zathura, unsigned int page_id)
     goto error_out;
   }
 
-  /* render page */
   zathura_page_t* page = zathura_document_get_page(zathura->document, page_id);
-
   if (page == NULL) {
     goto error_out;
   }
 
   zathura_document_set_current_page_number(zathura->document, page_id);
-  zathura->global.update_page_number = false;
 
-  page_offset_t offset;
-  page_calculate_offset(zathura, page, &offset);
-
-  GtkAdjustment* view_vadjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-  GtkAdjustment* view_hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-  zathura_adjustment_set_value(view_hadjustment, offset.x);
-  zathura_adjustment_set_value(view_vadjustment, offset.y);
-
-  statusbar_page_number_update(zathura);
-
-  return true;
+  /* negative position means auto */
+  return position_set(zathura, -1, -1);
 
 error_out:
-
   return false;
 }
 
@@ -1034,6 +1013,22 @@ statusbar_page_number_update(zathura_t* zathura)
   if (zathura->document != NULL) {
     char* page_number_text = g_strdup_printf("[%d/%d]", current_page_number + 1, number_of_pages);
     girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.page_number, page_number_text);
+
+    bool page_number_in_window_title = false;
+    girara_setting_get(zathura->ui.session, "window-title-page", &page_number_in_window_title);
+
+    if (page_number_in_window_title == true) {
+      bool basename_only = false;
+      girara_setting_get(zathura->ui.session, "window-title-basename", &basename_only);
+      char* title = g_strdup_printf("%s %s",
+        (basename_only == true)
+          ? zathura_document_get_basename(zathura->document)
+          : zathura_document_get_path(zathura->document),
+        page_number_text);
+      girara_set_window_title(zathura->ui.session, title);
+      g_free(title);
+    }
+
     g_free(page_number_text);
   } else {
     girara_statusbar_item_set_text(zathura->ui.session, zathura->ui.statusbar.page_number, "");
@@ -1041,7 +1036,8 @@ statusbar_page_number_update(zathura_t* zathura)
 }
 
 void
-page_widget_set_mode(zathura_t* zathura, unsigned int pages_per_row, unsigned int first_page_column)
+page_widget_set_mode(zathura_t* zathura, unsigned int page_padding,
+                     unsigned int pages_per_row, unsigned int first_page_column)
 {
   /* show at least one page */
   if (pages_per_row == 0) {
@@ -1052,7 +1048,6 @@ page_widget_set_mode(zathura_t* zathura, unsigned int pages_per_row, unsigned in
   if (first_page_column < 1) {
     first_page_column = 1;
   }
-
   if (first_page_column > pages_per_row) {
     first_page_column = ((first_page_column - 1) % pages_per_row) + 1;
   }
@@ -1064,9 +1059,18 @@ page_widget_set_mode(zathura_t* zathura, unsigned int pages_per_row, unsigned in
   gtk_container_foreach(GTK_CONTAINER(zathura->ui.page_widget), remove_page_from_table, (gpointer)0);
 
   unsigned int number_of_pages     = zathura_document_get_number_of_pages(zathura->document);
+
 #if (GTK_MAJOR_VERSION == 3)
+  gtk_grid_set_row_spacing(GTK_GRID(zathura->ui.page_widget), page_padding);
+  gtk_grid_set_column_spacing(GTK_GRID(zathura->ui.page_widget), page_padding);
+
 #else
-  gtk_table_resize(GTK_TABLE(zathura->ui.page_widget), ceil((number_of_pages + first_page_column - 1) / pages_per_row), pages_per_row);
+  gtk_table_set_row_spacings(GTK_TABLE(zathura->ui.page_widget), page_padding);
+  gtk_table_set_col_spacings(GTK_TABLE(zathura->ui.page_widget), page_padding);
+
+  unsigned int ncol = pages_per_row;
+  unsigned int nrow = (number_of_pages + first_page_column - 1 + ncol - 1) / ncol;
+  gtk_table_resize(GTK_TABLE(zathura->ui.page_widget), nrow, ncol);
 #endif
 
   for (unsigned int i = 0; i < number_of_pages; i++) {
@@ -1085,44 +1089,127 @@ page_widget_set_mode(zathura_t* zathura, unsigned int pages_per_row, unsigned in
   gtk_widget_show_all(zathura->ui.page_widget);
 }
 
-static gboolean
-position_set_delayed_impl(gpointer data)
+bool
+position_set(zathura_t* zathura, double position_x, double position_y)
 {
-  position_set_delayed_t* p = (position_set_delayed_t*) data;
-
-  GtkScrolledWindow *window = GTK_SCROLLED_WINDOW(p->zathura->ui.session->gtk.view);
-  GtkAdjustment* vadjustment = gtk_scrolled_window_get_vadjustment(window);
-  GtkAdjustment* hadjustment = gtk_scrolled_window_get_hadjustment(window);
-
-  /* negative values mean: don't set the position */
-  if (p->position_x >= 0) {
-    zathura_adjustment_set_value(hadjustment, p->position_x);
+  if (zathura == NULL || zathura->document == NULL) {
+    goto error_out;
   }
 
-  if (p->position_y >= 0) {
-    zathura_adjustment_set_value(vadjustment, p->position_y);
+  double comppos_x, comppos_y;
+  unsigned int page_id = zathura_document_get_current_page_number(zathura->document);
+
+  /* xalign = 0.5: center horizontally (with the page, not the document) */
+  /* yalign = 0.0: align page an viewport edges at the top               */
+  page_number_to_position(zathura->document, page_id, 0.5, 0.0, &comppos_x, &comppos_y);
+
+  /* automatic horizontal adjustment */
+  zathura_adjust_mode_t adjust_mode = zathura_document_get_adjust_mode(zathura->document);
+
+  /* negative position_x mean: use the computed value */
+  if (position_x < 0) {
+    position_x = comppos_x;
+    bool zoom_center = false;
+    girara_setting_get(zathura->ui.session, "zoom-center", &zoom_center);
+
+    /* center horizontally */
+    if (adjust_mode == ZATHURA_ADJUST_BESTFIT ||
+      adjust_mode == ZATHURA_ADJUST_WIDTH ||
+      zoom_center) {
+        position_x = 0.5;
+    }
   }
 
-  g_free(p);
+  if (position_y < 0) {
+    position_y = comppos_y;
+  }
 
-  return FALSE;
+  /* set the position */
+  zathura_document_set_position_x(zathura->document, position_x);
+  zathura_document_set_position_y(zathura->document, position_y);
+
+  /* trigger a 'change' event for both adjustments */
+  refresh_view(zathura);
+
+  return true;
+
+error_out:
+  return false;
 }
 
+
+void
+refresh_view(zathura_t* zathura) {
+  g_return_if_fail(zathura != NULL);
+
+  /* emit a custom refresh-view signal */
+  g_signal_emit(zathura->ui.session->gtk.view, zathura->signals.refresh_view,
+                0, zathura);
+}
+
+
 bool
-position_set_delayed(zathura_t* zathura, double position_x, double position_y)
-{
-  position_set_delayed_t* p = g_malloc0(sizeof(position_set_delayed_t));
+adjust_view(zathura_t* zathura) {
+    g_return_val_if_fail(zathura != NULL, false);
 
-  p->zathura    = zathura;
-  p->position_x = position_x;
-  p->position_y = position_y;
+  if (zathura->ui.page_widget == NULL || zathura->document == NULL) {
+    goto error_ret;
+  }
 
-  gdk_threads_add_idle(position_set_delayed_impl, p);
+  zathura_adjust_mode_t adjust_mode = zathura_document_get_adjust_mode(zathura->document);
+  if (adjust_mode == ZATHURA_ADJUST_NONE) {
+    /* there is nothing todo */
+    goto error_ret;
+  }
 
-  return FALSE;
+  unsigned int cell_height = 0, cell_width = 0;
+  unsigned int document_height = 0, document_width = 0;
+  unsigned int view_height = 0, view_width = 0;
+
+  zathura_document_get_cell_size(zathura->document, &cell_height, &cell_width);
+  zathura_document_get_document_size(zathura->document, &document_height, &document_width);
+  zathura_document_get_viewport_size(zathura->document, &view_height, &view_width);
+
+  if (view_height == 0 || view_width == 0 || cell_height == 0 || cell_width == 0) {
+    goto error_ret;
+  }
+
+  double page_ratio = (double)cell_height / (double)document_width;
+  double view_ratio = (double)view_height / (double)view_width;
+  double scale = zathura_document_get_scale(zathura->document);
+  double newscale = scale;
+
+  if (adjust_mode == ZATHURA_ADJUST_WIDTH ||
+      (adjust_mode == ZATHURA_ADJUST_BESTFIT && page_ratio < view_ratio)) {
+    newscale = scale * (double)view_width / (double)document_width;
+
+  } else if (adjust_mode == ZATHURA_ADJUST_BESTFIT) {
+    newscale = scale * (double)view_height / (double)cell_height;
+
+  } else {
+    goto error_ret;
+  }
+
+  /* save new scale and recompute cell size */
+  zathura_document_set_scale(zathura->document, newscale);
+  unsigned int new_cell_height = 0, new_cell_width = 0;
+  zathura_document_get_cell_size(zathura->document, &new_cell_height, &new_cell_width);
+
+  /* if the change in scale changes page cell dimensions by at least one pixel, render */
+  if (abs(new_cell_width - cell_width) > 1 ||
+      abs(new_cell_height - cell_height) > 1) {
+    render_all(zathura);
+    refresh_view(zathura);
+
+ /* otherwise set the old scale and leave */
+  } else {
+    zathura_document_set_scale(zathura->document, scale);
+  }
+
+error_ret:
+  return false;
 }
 
-
 bool
 zathura_jumplist_has_previous(zathura_t* zathura)
 {
@@ -1130,7 +1217,7 @@ zathura_jumplist_has_previous(zathura_t* zathura)
 }
 
 bool
-zathura_jumplist_has_has_next(zathura_t* zathura)
+zathura_jumplist_has_next(zathura_t* zathura)
 {
   return girara_list_iterator_has_next(zathura->jumplist.cur);
 }
@@ -1161,184 +1248,128 @@ zathura_jumplist_backward(zathura_t* zathura)
   }
 }
 
-void
+static void
+zathura_jumplist_reset_current(zathura_t* zathura)
+{
+  g_return_if_fail(zathura != NULL && zathura->jumplist.cur != NULL);
+
+  while (true) {
+    if (girara_list_iterator_has_next(zathura->jumplist.cur) == false) {
+      return;
+    }
+
+    girara_list_iterator_next(zathura->jumplist.cur);
+  }
+}
+
+static void
 zathura_jumplist_append_jump(zathura_t* zathura)
 {
+  g_return_if_fail(zathura != NULL && zathura->jumplist.list != NULL);
+
   zathura_jump_t *jump = g_malloc(sizeof(zathura_jump_t));
-  if (jump != NULL) {
-    jump->page = 0;
-    jump->x = 0;
-    jump->y = 0;
+  jump->page = 0;
+  jump->x = 0.0;
+  jump->y = 0.0;
+  girara_list_append(zathura->jumplist.list, jump);
 
-    /* remove right tail after current */
-    if (zathura->jumplist.cur != NULL) {
-      girara_list_iterator_t *it = girara_list_iterator_copy(zathura->jumplist.cur);
-      girara_list_iterator_next(it);
-      while (girara_list_iterator_is_valid(it)) {
-        girara_list_iterator_remove(it);
-        zathura->jumplist.size = zathura->jumplist.size - 1;
-      }
-      g_free(it);
+  if (zathura->jumplist.size == 0) {
+    zathura->jumplist.cur = girara_list_iterator(zathura->jumplist.list);
+  }
+
+  ++zathura->jumplist.size;
+  zathura_jumplist_trim(zathura);
+}
+
+void
+zathura_jumplist_trim(zathura_t* zathura)
+{
+  g_return_if_fail(zathura != NULL && zathura->jumplist.list != NULL && zathura->jumplist.size != 0);
+
+  girara_list_iterator_t* cur = girara_list_iterator(zathura->jumplist.list);
+
+  while (zathura->jumplist.size > zathura->jumplist.max_size) {
+    if (girara_list_iterator_data(cur) == girara_list_iterator_data(zathura->jumplist.cur)) {
+      girara_list_iterator_free(zathura->jumplist.cur);
+      zathura->jumplist.cur = NULL;
     }
 
-    /* trim from beginning until max_size */
-    girara_list_iterator_t *it = girara_list_iterator(zathura->jumplist.list);
-    while (zathura->jumplist.size >= zathura->jumplist.max_size && girara_list_iterator_is_valid(it)) {
-      girara_list_iterator_remove(it);
-      zathura->jumplist.size = zathura->jumplist.size - 1;
-    }
-    g_free(it);
+    girara_list_iterator_remove(cur);
+    --zathura->jumplist.size;
+  }
 
-    girara_list_append(zathura->jumplist.list, jump);
-    zathura->jumplist.size = zathura->jumplist.size + 1;
+  if (zathura->jumplist.size == 0 || (zathura->jumplist.size != 0 && zathura->jumplist.cur != NULL)) {
+    girara_list_iterator_free(cur);
+  } else {
+    zathura->jumplist.cur = cur;
   }
 }
 
 void
 zathura_jumplist_add(zathura_t* zathura)
 {
-  if (zathura->jumplist.list != NULL) {
-    unsigned int pagenum = zathura_document_get_current_page_number(zathura->document);
-    zathura_jump_t* cur = zathura_jumplist_current(zathura);
-    if (cur && cur->page == pagenum) {
-      return;
-    }
+  g_return_if_fail(zathura != NULL && zathura->document != NULL && zathura->jumplist.list != NULL);
 
-    zathura_jumplist_append_jump(zathura);
-    girara_list_iterator_next(zathura->jumplist.cur);
-    zathura_jumplist_save(zathura);
+  unsigned int pagenum = zathura_document_get_current_page_number(zathura->document);
+  double x = zathura_document_get_position_x(zathura->document);
+  double y = zathura_document_get_position_y(zathura->document);
+
+  if (zathura->jumplist.size != 0) {
+    zathura_jumplist_reset_current(zathura);
+
+    zathura_jump_t* cur = zathura_jumplist_current(zathura);
+
+    if (cur != NULL) {
+      if (cur->page == pagenum && cur->x == x && cur->y == y) {
+        return;
+      }
+    }
   }
+
+  zathura_jumplist_append_jump(zathura);
+  zathura_jumplist_reset_current(zathura);
+  zathura_jumplist_save(zathura);
 }
 
+bool
+zathura_jumplist_load(zathura_t* zathura, const char* file)
+{
+  g_return_val_if_fail(zathura != NULL && zathura->database != NULL && file != NULL, false);
 
-void
+  zathura->jumplist.list = zathura_db_load_jumplist(zathura->database, file);
+
+  if (zathura->jumplist.list == NULL) {
+    girara_error("Failed to load the jumplist from the database");
+
+    return false;
+  }
+
+  zathura->jumplist.size = girara_list_size(zathura->jumplist.list);
+
+  if (zathura->jumplist.size != 0) {
+    zathura->jumplist.cur = girara_list_iterator(zathura->jumplist.list);
+    zathura_jumplist_reset_current(zathura);
+    zathura_jumplist_trim(zathura);
+    girara_debug("Loaded the jumplist from the database");
+  } else {
+    girara_debug("No jumplist for this file in the database yet");
+  }
+
+  return true;
+}
+
+static void
 zathura_jumplist_save(zathura_t* zathura)
 {
+  g_return_if_fail(zathura != NULL && zathura->document != NULL);
+
   zathura_jump_t* cur = zathura_jumplist_current(zathura);
 
   unsigned int pagenum = zathura_document_get_current_page_number(zathura->document);
 
   if (cur) {
-    /* get position */
-    GtkAdjustment* view_vadjustment = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-    GtkAdjustment* view_hadjustment = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(zathura->ui.session->gtk.view));
-
     cur->page = pagenum;
-    cur->x = gtk_adjustment_get_value(view_hadjustment) / zathura_document_get_scale(zathura->document);
-    cur->y = gtk_adjustment_get_value(view_vadjustment) / zathura_document_get_scale(zathura->document);;
+    cur->x = zathura_document_get_position_x(zathura->document);
+    cur->y = zathura_document_get_position_y(zathura->document);
   }
 }
-
-static bool
-zathura_page_cache_is_cached(zathura_t* zathura, unsigned int page_index)
-{
-  g_return_val_if_fail(zathura != NULL, false);
-
-  unsigned int i;
-
-  if (zathura->page_cache.num_cached_pages != 0) {
-    for (i = 0; i < zathura->page_cache.size; ++i) {
-      if (zathura->page_cache.cache[i] >= 0 && page_index == (unsigned int)zathura->page_cache.cache[i]) {
-        girara_debug("Page %d is a cache hit", page_index + 1);
-        return true;
-      }
-    }
-  }
-
-  girara_debug("Page %d is a cache miss", page_index + 1);
-  return false;
-}
-
-static ssize_t
-zathura_page_cache_lru_invalidate(zathura_t* zathura)
-{
-   g_return_val_if_fail(zathura != NULL, -1);
-
-   ssize_t lru_index = 0;
-   gint64 view_time = 0;
-   gint64 lru_view_time = G_MAXINT64;
-   GtkWidget* page_widget;
-
-   for (unsigned int i = 0; i < zathura->page_cache.size; ++i) {
-     page_widget = zathura_page_get_widget(zathura, zathura_document_get_page(zathura->document, zathura->page_cache.cache[i]));
-     g_return_val_if_fail(page_widget != NULL, -1);
-     g_object_get(G_OBJECT(page_widget), "last-view", &view_time, NULL);
-
-     if (view_time < lru_view_time) {
-       lru_view_time = view_time;
-       lru_index = i;
-     }
-   }
-
-   zathura_page_t* page = zathura_document_get_page(zathura->document, zathura->page_cache.cache[lru_index]);
-   g_return_val_if_fail(page != NULL, -1);
-
-   page_widget = zathura_page_get_widget(zathura, page);
-   g_return_val_if_fail(page_widget != NULL, -1);
-
-   zathura_page_widget_update_surface(ZATHURA_PAGE(page_widget), NULL);
-   girara_debug("Invalidated page %d at cache index %ld", zathura->page_cache.cache[lru_index] + 1, lru_index);
-   zathura->page_cache.cache[lru_index] = -1;
-   --zathura->page_cache.num_cached_pages;
-
-   return lru_index;
-}
-
-static bool
-zathura_page_cache_is_full(zathura_t* zathura, bool* result)
-{
-  g_return_val_if_fail(zathura != NULL && result != NULL, false);
-
-  *result = zathura->page_cache.num_cached_pages == zathura->page_cache.size;
-
-  return true;
-}
-
-void
-zathura_page_cache_invalidate_all(zathura_t* zathura)
-{
-  g_return_if_fail(zathura != NULL);
-
-  unsigned int i;
-
-  for (i = 0; i < zathura->page_cache.size; ++i) {
-    zathura->page_cache.cache[i] = -1;
-  }
-
-  zathura->page_cache.num_cached_pages = 0;
-}
-
-void
-zathura_page_cache_add(zathura_t* zathura, unsigned int page_index)
-{
-  g_return_if_fail(zathura != NULL);
-
-  zathura_page_t* page = zathura_document_get_page(zathura->document, page_index);
-
-  g_return_if_fail(page != NULL);
-
-  if (zathura_page_cache_is_cached(zathura, page_index)) {
-    return;
-  }
-
-  bool full;
-
-  if (zathura_page_cache_is_full(zathura, &full) == false) {
-    return;
-  } else if (full == true) {
-    ssize_t idx = zathura_page_cache_lru_invalidate(zathura);
-
-    if (idx == -1) {
-      return;
-    }
-
-    zathura->page_cache.cache[idx] = page_index;
-    ++zathura->page_cache.num_cached_pages;
-    girara_debug("Page %d is cached at cache index %ld", page_index + 1, idx);
-    return;
-  }
-
-  zathura->page_cache.cache[zathura->page_cache.num_cached_pages++] = page_index;
-  girara_debug("Page %d is cached at cache index %d", page_index + 1, zathura->page_cache.num_cached_pages - 1);
-  return;
-}
